/*
 * AbstractPathWriter.cs
 *
 * Bread Crumb Trail Control
 * Version: 2.0
 *
 * Copyright (c) 2004, 2006, TopCoder, Inc. All rights reserved
 *
 * Authors: elefth, Standlove
 *          Pops, TCSDEVELOPER
 */

using System;
using System.Collections;
using System.Web;
using System.Web.UI;

namespace TopCoder.Web.UI.WebControl.BreadCrumb
{
    /// <summary>
    /// An abstract implementation of the <see cref="IPathWriter"/> which provides
    /// some of the common functionality when writing a path.
    /// </summary>
    ///
    /// <remarks>
    /// This abstract implementation will provide the following functionality:
    /// <list type="bullet">
    /// <term>
    /// Provides three methods: <see cref="BeginPath"/>, <see cref="WriteNode"/>, and
    /// <see cref="EndPath"/> that will allow implementations to concentrate on the rendering
    /// and not the processing of the path.
    /// </term>
    /// <term>
    /// Processes the node list in the way the associated <see cref="BreadCrumbTrailControl"/> wants
    /// (ie. the PathDirection) and will call the various methods with the appropriate
    /// parameters (ie. root node style on the root, current node style on the current node,
    /// and node style on all others.
    /// </term>
    /// <term>
    /// The separator (if specified) will be written in-between the nodes automatically.
    /// </term>
    /// </list>
    /// It is strongly suggested that all implementations of the <see cref="IPathWriter"/> inherit
    /// from this abstract base to not duplicate the processing of the path and to 'future' proof
    /// the implementations against future negligable changes.
    /// </remarks>
    ///
    /// <threadsafety static="true" instance="true"/>
    public abstract class AbstractPathWriter : IPathWriter
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="AbstractPathWriter"/> class.
        /// </summary>
        protected AbstractPathWriter()
        {
            // do nothing
        }

        /// <summary>
        /// Processes the nodes by calling the various methods for subclasses to override
        /// to perform the formatting/writing of the path.
        /// </summary>
        ///
        /// <remarks>
        /// This method will:
        /// <list type="number">
        /// <term>
        /// Call the <see cref="BeginPath"/> method to allow initialization logic to be performed.
        /// </term>
        /// <term>
        /// Get the ErrorDisposition from the parent control. If the ErrorDisposition is SilentlyIgnore,
        /// any exceptions in the following logic will be ignored and the logic will continue.
        /// If the ErrorDisposition is ThrowException, any exception encountered will be wrapped
        /// in a <see cref="BreadCrumbException"/> (unless it's already a <see cref="BreadCrumbException"/>)
        /// and returned immediately (in a 1 length array). If it's RenderException, all exceptions
        /// will be stored in a local ArrayList and returned out at the end.
        /// </term>
        /// <term>
        /// Get the PathDirection from the parent control.
        /// </term>
        /// <term>
        /// If PathDirection is RootToCurrent, iterate the nodes from 0 index to length - 1.
        /// If PathDirection is CurrentToRoot, iterate the nodes from length - 1 to 0.
        /// </term>
        /// <term>
        /// If the index position is 0, the node style will be the parent.RootNodeStyle.
        /// If the root node style is null, parent.NodeStyle is used instead.
        /// </term>
        /// <term>
        /// If the index position is length - 1, the node style will be the parent.CurrentNodeStyle.
        /// If the current node style is null, parent.NodeStyle is used instead.
        /// </term>
        /// <term>
        /// If the index position is any other position, the node style will be the parent.NodeStyle.
        /// If only one node to write, parent.RootNodeStyle will be used if available.
        /// </term>
        /// <term>
        /// Call <see cref="WriteNode"/> to write out the node.
        /// </term>
        /// <term>
        /// If we aren't on the last element in the iteration, write the separator via <see cref="WriteSeparator"/>.
        /// </term>
        /// <term>
        /// Once we are done iterating, call <see cref="EndPath"/> to perform any clean up code.
        /// </term>
        /// </list>
        /// </remarks>
        ///
        /// <param name="writer">
        /// A non-null writer to use.
        /// </param>
        /// <param name="nodes">
        /// The non-null, non-empty array of nodes to write.
        /// </param>
        /// <param name="parent">
        /// A non-null parent bread crumb trail control to get attributes from.
        /// </param>
        ///
        /// <returns>
        /// A null if no exceptions occur or a &gt;= 1 length array of exceptions that occurred.
        /// </returns>
        ///
        /// <exception cref="ArgumentNullException">
        /// Any of the argument is a null reference.
        /// </exception>
        /// <exception cref="ArgumentException">
        /// <paramref name="nodes"/> is an empty array or contains a null/uninitialized element (the Url).
        /// </exception>
        public virtual BreadCrumbException[] WritePath(
            HtmlTextWriter writer, BreadCrumbNode[] nodes, BreadCrumbTrailControl parent)
        {
            if (writer == null)
            {
                throw new ArgumentNullException("writer", "'writer' cannot be null");
            }
            if (nodes == null)
            {
                throw new ArgumentNullException("nodes", "'nodes' cannot be null");
            }
            if (parent == null)
            {
                throw new ArgumentNullException("parent", "'parent' cannot be null");
            }
            if (nodes.Length == 0)
            {
                throw new ArgumentException("'nodes' cannot be an empty array", "nodes");
            }

            // prepare IList to accumulate exception and whether we should stop immediately if an error occurs
            ArrayList ex = new ArrayList();
            bool stopImmediate = (parent.ErrorDisposition == ErrorDisposition.ThrowException);

            try
            {
                // begin processing (default implementation does nothing)
                BeginPath(writer);
            }
            catch (Exception e)
            {
                // don't let any exception passthrough
                Helper.HandleException(e, "Error writing path in BeginPath", ex, parent.ErrorDisposition, false);

                // don't continue if not allowed
                if (stopImmediate)
                {
                    return ListToArray(ex);
                }
            }

            // determine last node position (depending direction)
            PathDirection dir = parent.PathDirection;
            int lastPos = ((dir == PathDirection.RootToCurrent) ? nodes.Length - 1 : 0);

            // iterate all nodes (using absolute position)
            for (int i = 0; i < nodes.Length; i++)
            {
                // determine actual node position to write (depending direction)
                int pos = ((dir == PathDirection.RootToCurrent) ? i : nodes.Length - 1 - i);

                // node's Url may not be initialized (indicating uninitialized node)
                if (nodes[pos].Url == null)
                {
                    throw new ArgumentException(string.Format(
                        "'nodes' contains null Url element at index {0}", pos), "nodes");
                }

                // atomically write a node
                try
                {
                    // determine the style and write the node
                    string nodeStyle = GetNodeStyle(pos, nodes.Length, parent);
                    WriteNode(writer, nodes[pos], nodeStyle, parent.MouseOverStyle);

                    // write separator except if we reached lastPos
                    if (pos != lastPos)
                    {
                        WriteSeparator(writer, parent);
                    }
                }
                catch (Exception e)
                {
                    // any error while processing a node (accumulate/ignore it)
                    Helper.HandleException(e, "Error writing node", ex, parent.ErrorDisposition, false);

                    // stop processing immediately if disposition is ThrowException
                    // otherwise we continue with next node
                    if (stopImmediate)
                    {
                        return ListToArray(ex);
                    }
                }
            }

            // end processing (default implementation does nothing)
            try
            {
                EndPath(writer);
            }
            catch (Exception e)
            {
                // don't let any exception passthrough
                Helper.HandleException(e, "Error writing path in EndPath", ex, parent.ErrorDisposition, false);
            }

            // return null or accumulated exceptions
            return ListToArray(ex);
        }

        /// <summary>
        /// This method is called prior to processing the node path and
        /// will allow subimplementations override to provide initialization logic.
        /// </summary>
        ///
        /// <remarks>
        /// Currently, other than argument checking, it does nothing.
        /// </remarks>
        ///
        /// <param name="writer">
        /// A non-null writer to write to.
        /// </param>
        ///
        /// <exception cref="ArgumentNullException">
        /// <paramref name="writer"/> is a null reference.
        /// </exception>
        protected virtual void BeginPath(HtmlTextWriter writer)
        {
            if (writer == null)
            {
                throw new ArgumentNullException("writer", "'writer' cannot be null");
            }
        }

        /// <summary>
        /// This method will be called once for each node in the path with the appropriate node style.
        /// </summary>
        ///
        /// <remarks>
        /// Subclasses will implement this abstract method to provide the processing for each node.
        /// </remarks>
        ///
        /// <param name="writer">
        /// A non-null writer to write to.
        /// </param>
        /// <param name="node">
        /// A non-null node to write.
        /// </param>
        /// <param name="nodeStyle">
        /// A style class to use for the node or null if no style should be used.
        /// </param>
        /// <param name="mouseOverStyle">
        /// The mouse over style class to use or null for none.
        /// </param>
        ///
        /// <exception cref="ArgumentNullException">
        /// <paramref name="writer"/> is null or <paramref name="node"/> has a null Url.
        /// </exception>
        protected abstract void WriteNode(
            HtmlTextWriter writer, BreadCrumbNode node, string nodeStyle, string mouseOverStyle);

        /// <summary>
        /// Writes the separator if specified.
        /// </summary>
        ///
        /// <remarks>
        /// This method will write PathSeparator if it's not null.
        /// The separator will be written enclosed in a &lt;span&gt; tag with
        /// a class attribute if PathSeparatorStyle is specified.
        /// </remarks>
        ///
        /// <param name="writer">
        /// A non-null writer to use.
        /// </param>
        /// <param name="parent">
        /// A non-null parent control to retrieve PathSeparator and PathSeparatorStyle.
        /// </param>
        ///
        /// <exception cref="ArgumentNullException">
        /// Any of the argument is a null reference.
        /// </exception>
        protected virtual void WriteSeparator(HtmlTextWriter writer, BreadCrumbTrailControl parent)
        {
            if (writer == null)
            {
                throw new ArgumentNullException("writer", "'writer' cannot be null");
            }
            if (parent == null)
            {
                throw new ArgumentNullException("parent", "'parent' cannot be null");
            }

            if (parent.PathSeparator == null)
            {
                // no path separator to write
                return;
            }

            // use <span> to add class attribute if a style is defined
            bool useSpan = (parent.PathSeparatorStyle != null);

            // add the style attribute and span
            if (useSpan)
            {
                writer.AddAttribute(HtmlTextWriterAttribute.Class, parent.PathSeparatorStyle);
                writer.RenderBeginTag(HtmlTextWriterTag.Span);
            }

            // render the separator itself (html encoded)
            writer.Write(HttpUtility.HtmlEncode(parent.PathSeparator));

            if (useSpan)
            {
                writer.RenderEndTag();
            }
        }

        /// <summary>
        /// This method is called post processing the node path and
        /// will allow subimplementations override to provide cleanup logic.
        /// </summary>
        ///
        /// <remarks>
        /// Currently, other than argument checking, it does nothing.
        /// </remarks>
        ///
        /// <param name="writer">
        /// A non-null writer to write to.
        /// </param>
        ///
        /// <exception cref="ArgumentNullException">
        /// <paramref name="writer"/> is a null reference.
        /// </exception>
        protected virtual void EndPath(HtmlTextWriter writer)
        {
            if (writer == null)
            {
                throw new ArgumentNullException("writer", "'writer' cannot be null");
            }
        }

        /// <summary>
        /// Gets the style for the node at the given position.
        /// </summary>
        ///
        /// <remarks>
        /// Position 0 will return RootNodeStyle (if it exists).
        /// Last position (length - 1) will return CurrentNodeStyle (if it exists).
        /// Any other position and condition will return NodeStyle.
        /// Note that if length is 1, RootNodeStyle is used (if it exists).
        /// </remarks>
        ///
        /// <param name="pos">
        /// The node position. 0 is the root node, length - 1 is the current node.
        /// </param>
        /// <param name="length">
        /// The total length of the nodes.
        /// </param>
        /// <param name="parent">
        /// The parent to retrieve style from.
        /// </param>
        ///
        /// <returns>
        /// A string representing the style for the node at the given position.
        /// </returns>
        private string GetNodeStyle(int pos, int length, BreadCrumbTrailControl parent)
        {
            // default is NodeStyle
            string style = parent.NodeStyle;

            // current position (or if length==1), use CurrentNodeStyle if it exists
            if ((pos == length - 1) && (parent.CurrentNodeStyle != null))
            {
                style = parent.CurrentNodeStyle;
            }

            // root position, use RootNodeStyle if it exists
            if ((pos == 0) && (parent.RootNodeStyle != null))
            {
                style = parent.RootNodeStyle;
            }

            return style;
        }

        /// <summary>
        /// Converts the given list to array.
        /// Used by WritePath to return accumulated exceptions.
        /// </summary>
        ///
        /// <param name="ex">
        /// The accumulated exceptions in an ArrayList.
        /// </param>
        ///
        /// <returns>
        /// The array of BreadCrumbException or null if array is empty.
        /// </returns>
        private BreadCrumbException[] ListToArray(ArrayList ex)
        {
            // return null or accumulated exceptions
            return ((ex.Count == 0) ? null : (BreadCrumbException[]) ex.ToArray(typeof(BreadCrumbException)));
        }
    }
}
