/*
 * INodeMatcher.cs
 *
 * Bread Crumb Trail Control
 * Version: 2.0
 *
 * Copyright (c) 2004, 2006, TopCoder, Inc. All rights reserved
 *
 * Authors: elefth, Standlove
 *          Pops, TCSDEVELOPER
 */

using System;

namespace TopCoder.Web.UI.WebControl.BreadCrumb.PathDiscovery
{
    /// <summary>
    /// Defines the contract for a node matcher.
    /// </summary>
    ///
    /// <remarks>
    /// <para>
    /// Implementations need to provide three functions:
    /// <list type="bullet">
    /// <item><term>
    /// A matching function that will take a node and determine if it matches.
    /// </term></item>
    /// <item><term>
    /// A default title string used for those nodes that match.
    /// </term></item>
    /// <item><term>
    /// A default url to use if no contextual information is available.
    /// </term></item>
    /// </list>
    /// Implementations of this inteface <b>must</b> override the Equals and GetHashCode methods
    /// because the implementation will be stored in an IDictionary.
    /// </para>
    /// <para>
    /// Implementations will be created by <see cref="ISiteDiscovery"/> implementations
    /// when the site map is created. The <see cref="INodeMatcher"/> implementation will then be used
    /// by the <see cref="DynamicPathDiscovery"/> when it attempts to discover the path to the root node
    /// from some current node.
    /// It is also strongly suggested that implementations be marked with <see cref="SerializableAttribute"/>.
    /// </para>
    /// <para>
    /// <b>Thread Safety:</b> Implementations of this interface should be thread safe.
    /// </para>
    /// </remarks>
    public interface INodeMatcher
    {
        /// <summary>
        /// Returns the default title for any node that match this node matcher.
        /// </summary>
        ///
        /// <remarks>
        /// The default title can be used if the node has no title assigned to it.
        /// </remarks>
        ///
        /// <value>
        /// A non-null, non-empty (trimmed) string representing the default title.
        /// </value>
        string Title
        {
            get;
        }

        /// <summary>
        /// Returns the default url that can be used if contextual node information is not available.
        /// </summary>
        ///
        /// <value>
        /// A non-null, non-empty (trimmed) string representing the url.
        /// </value>
        string Url
        {
            get;
        }

        /// <summary>
        /// Determines whether the specified node matches this node matcher.
        /// </summary>
        ///
        /// <param name="node">
        /// A non-null node to test.
        /// </param>
        ///
        /// <returns>
        /// true if the node matches, false otherwise.
        /// </returns>
        ///
        /// <exception cref="ArgumentNullException">
        /// <paramref name="node"/> is a null reference.
        /// </exception>
        bool Matches(BreadCrumbNode node);
    }
}


