/*
 * NodePair.cs
 *
 * Bread Crumb Trail Control
 * Version: 2.0
 *
 * Copyright (c) 2004, 2006, TopCoder, Inc. All rights reserved
 *
 * Authors: elefth, Standlove
 *          Pops, TCSDEVELOPER
 */

using System;

namespace TopCoder.Web.UI.WebControl.BreadCrumb.PathDiscovery
{
    /// <summary>
    /// This class is a typical pair class that holds a paired association of
    /// a node matcher to the matching node.
    /// </summary>
    ///
    /// <remarks>
    /// The node describes the actual node that matched the <see cref="NodeMatcher"/>
    /// and contains the contextual specific information.
    /// This class will be created by the <see cref="DynamicPathDiscovery"/> class as
    /// it's creating the current path. The <see cref="DynamicPathDiscovery"/> class will
    /// then use this class to hold the association between the contextual specific node and
    /// the matcher node that it matches. The <see cref="DynamicPathDiscovery"/> will then use
    /// this information to create the path from each of the path pairs.
    /// This class provides one additional benefit - the returned <see cref="BreadCrumbNode"/>
    /// will contain either the specified title (in the node itself) or if the title is null,
    /// the title from the <see cref="NodeMatcher"/> instead.
    /// </remarks>
    ///
    /// <threadsafety static="true" instance="true"/>
    [Serializable]
    internal class NodePair
    {
        /// <summary>
        /// Represents the actual node that holds the contextual information (query string, etc)
        /// that matches the node matcher. This variable is set by the constructor,
        /// is immutable and can never be null.
        /// The node will be referenced in the associated getter method.
        /// </summary>
        private readonly BreadCrumbNode node;

        /// <summary>
        /// Represents the node matcher that holds the matching information and the default title.
        /// This variable is set by the constructor, is immutable and can be null
        /// (if there is no associated nodeMatcher).
        /// The path node will be referenced in the associated getter method and the Node getter method.
        /// </summary>
        private readonly INodeMatcher nodeMatcher;

        /// <summary>
        /// Gets the node, containing the contextual information, that matches the corresponding
        /// <see cref="NodeMatcher"/>.
        /// </summary>
        ///
        /// <remarks>
        /// The returned node will have its title taken from the actual node/page if it's not null.
        /// Otherwise the default title from the <see cref="NodeMatcher"/> is used (if there is one).
        /// If no node matcher is associated, the returned node will have a null title.
        /// </remarks>
        ///
        /// <value>
        /// The node, containing the contextual information, that matches the corresponding
        /// <see cref="NodeMatcher"/>.
        /// </value>
        public BreadCrumbNode Node
        {
            get
            {
                // return node if Title is not null
                if (node.Title != null)
                {
                    return node;
                }

                // use NodeMatcher's Title (if there's a nodeMatcher) or null if no NodeMatcher
                string title = ((NodeMatcher == null) ? null : NodeMatcher.Title);
                return new BreadCrumbNode(title, node.Url);
            }
        }

        /// <summary>
        /// Getter the node matcher associated with this <see cref="Node"/>.
        /// </summary>
        ///
        /// <value>
        /// The node matcher associated with this <see cref="Node"/> which could be null.
        /// </value>
        public INodeMatcher NodeMatcher
        {
            get
            {
                return nodeMatcher;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NodePair"/> class
        /// with the specified node and matcher.
        /// </summary>
        ///
        /// <param name="node">
        /// The bread crumb node that matches the path node.
        /// </param>
        /// <param name="nodeMatcher">
        /// The possibly null node matcher.
        /// </param>
        ///
        /// <exception cref="ArgumentNullException">
        /// <paramref name="node"/> has a null Url.
        /// </exception>
        internal NodePair(BreadCrumbNode node, INodeMatcher nodeMatcher)
        {
            if (node.Url == null)
            {
                throw new ArgumentNullException("node", "'node.Url' cannot be null");
            }

            this.node = node;
            this.nodeMatcher = nodeMatcher;
        }
    }
}
