/*
 * XmlPathDiscovery.cs
 *
 * Bread Crumb Trail Control
 * Version: 2.0
 *
 * Copyright (c) 2004, 2006, TopCoder, Inc. All rights reserved
 *
 * Authors: elefth, Standlove
 *          Pops, TCSDEVELOPER
 */

using System;
using System.Collections;
using System.IO;
using System.Web.UI;
using System.Xml;
using System.Xml.Schema;
using TopCoder.Web.UI.WebControl.BreadCrumb.PathDiscovery.NodeMatcher;

namespace TopCoder.Web.UI.WebControl.BreadCrumb.PathDiscovery
{
    /// <summary>
    /// An implementation of the <see cref="IPathDiscovery"/> and <see cref="ISiteDiscovery"/>
    /// which creates a static site map from either a filename or string source
    /// (both of which is an XML document).
    /// </summary>
    ///
    /// <remarks>
    /// <para>
    /// This class will parse the document to create a static map of regex matchers
    /// (where each node can provide a regex matching). If used as a <see cref="IPathDiscovery"/>,
    /// this provides a static view of the site (ie. no dynamic capabilities).
    /// </para>
    /// <para>
    /// This implementation will be used as either a non-dynamic path discovery operation
    /// to be used by the <see cref="BreadCrumbTrailControl"/> or can be used as
    /// an <see cref="ISiteDiscovery"/> for any other <see cref="IPathDiscovery"/> implementation
    /// (the <see cref="DynamicPathDiscovery"/> will use it as the source of site information).
    /// The <see cref="BreadCrumbTrailControl"/> will either construct this class directly
    /// (with a path override argument) or automatically via the object factory.
    /// This class will store the static site map to a session variable called <see cref="CURRENT_SITEMAP"/>
    /// so that the file/string source needs to be parsed only once.
    /// A reverse site map is also internally managed to enable faster path finding to the root node.
    /// </para>
    /// </remarks>
    ///
    /// <threadsafety static="true" instance="true"/>
    public class XmlPathDiscovery : IPathDiscovery, ISiteDiscovery
    {
        /// <summary>
        /// Defines the Session scope attribute name that will contain the site map.
        /// </summary>
        ///
        /// <remarks>
        /// This constant will be used by this class to retrieve the site map.
        /// The site map will be an <see cref="IDictionary"/> with <see cref="INodeMatcher"/>
        /// as the key and an <see cref="IList"/> of <see cref="INodeMatcher"/> as the value and can be empty.
        /// This <see cref="IDictionary"/> map represents the <see cref="INodeMatcher"/> parent to its children
        /// <see cref="INodeMatcher"/>(s) if any. The list can be empty if there are no children of the parent
        /// (ie. a leaf node).
        /// </remarks>
        public static readonly string CURRENT_SITEMAP = typeof(XmlPathDiscovery).FullName + ".siteMap";

        /// <summary>
        /// Defines the Session scope attribute name that will contain the reverse site map mapping.
        /// </summary>
        ///
        /// <remarks>
        /// Similar to <see cref="CURRENT_SITEMAP"/>, however the IDictionary will contain
        /// a mapping from children to multiple parent.
        /// The root node will have an empty value list.
        /// </remarks>
        public static readonly string CURRENT_SITEMAP_REVERSE = CURRENT_SITEMAP + ".reverse";

        /// <summary>
        /// The Xml Schema for a site map xml file.
        /// Used to validate the loaded xml.
        /// </summary>
        private const string BreadCrumbTrailSchema =
            @"<?xml version=""1.0"" encoding=""UTF-8"" ?>
            <xsd:schema xmlns:xsd=""http://www.w3.org/2001/XMLSchema"">
                <xsd:element name=""BreadCrumb"" type=""BreadCrumb""/>
                <xsd:complexType name=""BreadCrumb"">
                    <xsd:sequence>
                        <xsd:element name=""bc_node"" type=""bc_node"" minOccurs=""0"" maxOccurs=""unbounded""/>
                    </xsd:sequence>
                    <xsd:attribute name=""version"" type=""xsd:token""/>
                    <xsd:attribute name=""noNamespaceSchemaLocation"" type=""xsd:token""/>
                </xsd:complexType>
                <xsd:complexType name=""bc_node"">
                    <xsd:sequence>
                        <xsd:element minOccurs=""0"" maxOccurs=""unbounded"" name=""bc_node"" type=""bc_node""/>
                    </xsd:sequence>
                    <xsd:attribute name=""pattern"" type=""xsd:token"" use=""required""/>
                    <xsd:attribute name=""title"" type=""xsd:token"" use=""required""/>
                    <xsd:attribute name=""url"" type=""xsd:token"" use=""required""/>
                </xsd:complexType>
            </xsd:schema>";

        /// <summary>
        /// The element tag for a bread crumb node.
        /// </summary>
        private const string BcNodeTag = "bc_node";

        /// <summary>
        /// The attribute name for pattern.
        /// </summary>
        private const string PatternAttr = "pattern";

        /// <summary>
        /// The attribute name for title.
        /// </summary>
        private const string TitleAttr = "title";

        /// <summary>
        /// The attribute name for url.
        /// </summary>
        private const string UrlAttr = "url";

        /// <summary>
        /// Represents the source of the site data.
        /// The source will be set by the constructor, is immutable and will never be null
        /// but may be an empty (trim'd) string.
        /// This variable is referenceable by the GetSiteMap method.
        /// Note that this source may need to be re-mapped according to the server page context.
        /// </summary>
        private readonly string source;

        /// <summary>
        /// Represents whether the source variable represents a file (true) or a raw xml string (false).
        /// The isFile will be set by the constructor and is immutable.
        /// This variable is referencable by the GetSiteMap method.
        /// </summary>
        private readonly bool isFile;

        /// <summary>
        /// Initializes a new instance of the <see cref="XmlPathDiscovery"/> class
        /// with the specified filename as the site map source.
        /// </summary>
        ///
        /// <param name="filename">
        /// The Xml filename containing the site map.
        /// </param>
        ///
        /// <exception cref="ArgumentNullException">
        /// <paramref name="filename"/> is a null reference.
        /// </exception>
        /// <exception cref="ArgumentException">
        /// <paramref name="filename"/> is an empty trimmed string.
        /// </exception>
        public XmlPathDiscovery(string filename)
            : this(filename, true)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="XmlPathDiscovery"/> class
        /// with the specified source string or source file name.
        /// </summary>
        ///
        /// <param name="source">
        /// The source string or the source file name.
        /// </param>
        /// <param name="isFile">
        /// Whether <paramref name="source"/> is a file (true) or a string (false).
        /// </param>
        ///
        /// <exception cref="ArgumentNullException">
        /// <paramref name="source"/> is a null reference.
        /// </exception>
        /// <exception cref="ArgumentException">
        /// <paramref name="source"/> is an empty trimmed string.
        /// </exception>
        public XmlPathDiscovery(string source, bool isFile)
        {
            if (source == null)
            {
                throw new ArgumentNullException("source", "'source' cannot be null");
            }
            if (source.Trim().Length == 0)
            {
                throw new ArgumentException("'source' cannot be an empty trimmed string", "source");
            }

            this.source = source;
            this.isFile = isFile;
        }

        /// <summary>
        /// Returns an <see cref="IDictionary"/> of <see cref="INodeMatcher"/> keys with
        /// an <see cref="IList"/> of <see cref="INodeMatcher"/> elements as the value.
        /// </summary>
        ///
        /// <remarks>
        /// The key <see cref="INodeMatcher"/> represents the parent and the value (the list)
        /// represents the child values. The value can be an empty list to represent
        /// no child elements (ie. it's a leaf node).
        /// Note that this method does the actual parsing of the xml source.
        /// It is advised that the returned IDictionary is cached by the caller.
        /// </remarks>
        ///
        /// <returns>
        /// An <see cref="IDictionary"/> of parent <see cref="INodeMatcher"/> to a list of
        /// <see cref="INodeMatcher"/> children.
        /// </returns>
        ///
        /// <exception cref="BreadCrumbException">
        /// Any error occurs parsing the data source.
        /// </exception>
        public IDictionary GetSiteMap()
        {
            // load & parse file/string source into XmlDocument
            XmlDocument doc = LoadSource();

            // recursively build the RegexNodeMatcher parent-children nodes
            IDictionary result = new Hashtable();
            BuildSiteMap(result, doc.DocumentElement);
            return result;
        }

        /// <summary>
        /// Determines the shortest path from the current node to the root
        /// based on the current site map.
        /// </summary>
        ///
        /// <remarks>
        /// This method will retrieve the site map using <see cref="GetSiteMap"/> and cache it
        /// in the page's Session with the key <see cref="CURRENT_SITEMAP"/>.
        /// Internally this method maintains a reverse site-map mapping in
        /// <see cref="CURRENT_SITEMAP_REVERSE"/>.
        /// A BFS algorithm is used to quickly find the first shortest path.
        /// In case several shortest paths were found, it would return
        /// the first path traced in the reverse site map (IDictionary hash and IList order).
        /// </remarks>
        ///
        /// <param name="currentNode">
        /// The current node.
        /// </param>
        /// <param name="page">
        /// The page context to save/get cached site map.
        /// </param>
        ///
        /// <returns>
        /// A path from root node (0th index) to current node ((n-1)th index)
        /// or null if no path can be found.
        /// </returns>
        ///
        /// <exception cref="ArgumentNullException">
        /// Any argument is null or currentNode contains null Url.
        /// </exception>
        /// <exception cref="BreadCrumbException">
        /// Any error occurs finding the path.
        /// </exception>
        public BreadCrumbNode[] GetPath(BreadCrumbNode currentNode, Page page)
        {
            if (currentNode.Url == null)
            {
                throw new ArgumentNullException("currentNode", "'currentNode.Url' cannot be null");
            }
            if (page == null)
            {
                throw new ArgumentNullException("page", "'page' cannot be null");
            }

            try
            {
                // get (reverse) site map from Session or create a new one
                IDictionary reverseSiteMap = GetReverseSiteMap(page);

                // result will be initialized to BreadCrumbNode[]
                // (representing the found shortest path in OnRootFound())
                object result = null;
                PathAlg.FindShortestToRoot(
                    currentNode, reverseSiteMap, new RootFound(OnRootFound), ref result);
                BreadCrumbNode[] shortestPath = (BreadCrumbNode[]) result;

                // did we find any shortest path?
                if (shortestPath != null)
                {
                    // replace last element with currentNode
                    shortestPath[shortestPath.Length - 1] = currentNode;
                }

                return shortestPath;
            }
            catch (BreadCrumbException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new BreadCrumbException(string.Format(
                    "Error retrieving path for url {0}", currentNode.Url), e);
            }
        }

        /// <summary>
        /// Loads the file/string source and parse/validate into XmlDocument.
        /// </summary>
        ///
        /// <returns>
        /// The XmlDocument representing the file/string source.
        /// </returns>
        ///
        /// <exception cref="BreadCrumbException">
        /// An error occurs while loading/parsing the source.
        /// </exception>
        private XmlDocument LoadSource()
        {
            XmlValidatingReader reader = null;
            try
            {
                XmlDocument doc = new XmlDocument();

                // create text reader (from file or string source)
                XmlTextReader txtReader = new XmlTextReader(
                    (isFile ? (TextReader) new StreamReader(source) : new StringReader(source)));

                // use validating reader (load the schema)
                reader = new XmlValidatingReader(txtReader);
                using (StringReader xmlSchema = new StringReader(BreadCrumbTrailSchema))
                {
                    reader.Schemas.Add(XmlSchema.Read(xmlSchema, null));
                }

                // load and validates the document
                doc.Load(reader);
                return doc;
            }
            catch (Exception e)
            {
                // quote the detail error message from the validator
                throw new BreadCrumbException(string.Format(
                    "Error loading/parsing the source: {0}", e.Message), e);
            }
            finally
            {
                if (reader != null)
                {
                    // automatically close all underlying streams
                    reader.Close();
                }
            }
        }

        /// <summary>
        /// Recursively build parent-children IDictionary entry.
        /// </summary>
        ///
        /// <remarks>
        /// The algorithm is: <pre>
        /// Retrieve all "bc_node" elements under startNode.
        /// For each node:
        /// 1. Create corresponding RegexNodeMatcher
        /// 2. Recursive call to process current Node (get IList of children)
        /// 3. Add to IDictionary (check first that node is not defined with children in several places)
        /// 4. Add to IList representing list of nodes underneath startNode
        /// return IList
        /// </pre>
        /// </remarks>
        ///
        /// <param name="result">
        /// The IDictionary to store parent-children (IList) nodes.
        /// </param>
        /// <param name="startNode">
        /// The current node to start processing from.
        /// </param>
        ///
        /// <returns>
        /// An IList containing all INodeMatcher instances
        /// of all bc_node tag underneath the startNode.
        /// </returns>
        ///
        /// <exception cref="BreadCrumbException">
        /// An error occurs parsing the source.
        /// </exception>
        private IList BuildSiteMap(IDictionary result, XmlNode startNode)
        {
            try
            {
                // get all "bc_node" elements under startNode
                XmlNodeList parentNodes = startNode.SelectNodes(BcNodeTag);
                IList parentList = new ArrayList();

                // process all these as parent nodes
                // if no bc_tag element underneath, this loop won't run and empty array is returned
                foreach (XmlNode parentNode in parentNodes)
                {
                    // create RegexNodeMatcher (key)
                    string title = parentNode.Attributes[TitleAttr].Value;
                    string url = parentNode.Attributes[UrlAttr].Value;
                    string pattern = parentNode.Attributes[PatternAttr].Value;
                    INodeMatcher parent = new RegexNodeMatcher(title, url, pattern);

                    // retrieve IList of child RegexNodeMatcher with this node as startNode
                    IList children = BuildSiteMap(result, parentNode);

                    // insert key and children into the result IDictionary
                    // but first, check whether we have seen this same node
                    if (result.Contains(parent))
                    {
                        // make sure that only one node can contain children, the other should be childless
                        IList prevChildren = (IList) result[parent];
                        int childMin = Math.Min(children.Count, prevChildren.Count);
                        int childMax = Math.Max(children.Count, prevChildren.Count);
                        if ((childMin > 0) && (childMax > 0))
                        {
                            throw new BreadCrumbException(string.Format(
                                "Node with title: {0} url: {1} pattern: {2} is defined in several places "
                                + "with children. Only one may define children, "
                                + "all other should be childless.",
                                title, url, pattern));
                        }

                        // we can safely overwrite previous entry if it was childless
                        if (prevChildren.Count == 0)
                        {
                            result[parent] = children;
                        }
                    }
                    else
                    {
                        // first seen
                        result[parent] = children;
                    }

                    // add to IList of nodes to be returned
                    parentList.Add(parent);
                }

                // return IList of nodes underneath startNode
                return parentList;
            }
            catch (BreadCrumbException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new BreadCrumbException("Error parsing site map xml", e);
            }
        }

        /// <summary>
        /// Gets the reverse site map from the Page's Session or
        /// create a new one from GetSiteMap.
        /// </summary>
        ///
        /// <param name="page">
        /// The page context to retrieve/save site map and its reverse site map.
        /// </param>
        ///
        /// <returns>
        /// The reverse site map.
        /// </returns>
        ///
        /// <exception cref="BreadCrumbException">
        /// Any error procesing the site map.
        /// </exception>
        private IDictionary GetReverseSiteMap(Page page)
        {
            try
            {
                // retrieve site map from the session
                object siteMap = page.Session[CURRENT_SITEMAP];
                if ((siteMap == null) || (!(siteMap is IDictionary)))
                {
                    // if not exist, create site map (and its reverse)
                    IDictionary newSiteMap = GetSiteMap();
                    page.Session[CURRENT_SITEMAP] = newSiteMap;
                    page.Session[CURRENT_SITEMAP_REVERSE] = PathAlg.CreateReverse(newSiteMap);
                }
                else
                {
                    // make sure a reverse site map exist
                    object reverseSiteMap = page.Session[CURRENT_SITEMAP_REVERSE];
                    if ((reverseSiteMap == null) || (!(reverseSiteMap is IDictionary)))
                    {
                        page.Session[CURRENT_SITEMAP_REVERSE] = PathAlg.CreateReverse((IDictionary) siteMap);
                    }
                }

                // return the reverse site map
                return (IDictionary) page.Session[CURRENT_SITEMAP_REVERSE];
            }
            catch (BreadCrumbException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new BreadCrumbException("Error processing sitemap", e);
            }
        }

        /// <summary>
        /// The delegate that is called by PathAlg.FindShortestToRoot
        /// when a path to root is found.
        /// This will create a BreadCrumbNode[] and copy the found path to the array
        /// in reverse order.
        /// </summary>
        ///
        /// <param name="path">
        /// IList of INodeMatcher representing the path to root
        /// (first element is the current node, last element is root node).
        /// </param>
        /// <param name="result">
        /// A reference to store the resulting array to.
        /// </param>
        ///
        /// <returns>
        /// Always false to indicate we only want first found shortest path.
        /// </returns>
        private bool OnRootFound(IList path, ref object result)
        {
            // copy the INodeMatchers in path (current->root) to result array (root->current)
            BreadCrumbNode[] shortestPath = new BreadCrumbNode[path.Count];
            for (int i = path.Count - 1, j = 0; i >= 0; i--, j++)
            {
                INodeMatcher matcher = (INodeMatcher) path[i];
                shortestPath[j] = new BreadCrumbNode(matcher.Title, matcher.Url);
            }
            result = shortestPath;
            return false;
        }
    }
}
