/*
 * TestRegexNodeMatcher.cs
 *
 * Bread Crumb Trail Control
 * Version: 2.0
 *
 * Copyright (c) 2004, 2006, TopCoder, Inc. All rights reserved
 *
 * Authors: Standlove, TCSDEVELOPER
 */

using System;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using NUnit.Framework;
using TopCoder.Web.UI.WebControl.BreadCrumb.PathDiscovery.Tests;

namespace TopCoder.Web.UI.WebControl.BreadCrumb.PathDiscovery.NodeMatcher.Tests
{
    /// <summary>
    /// Unit tests for the <see cref="RegexNodeMatcher"/> class.
    /// </summary>
    [TestFixture]
    public class TestRegexNodeMatcher
    {
        /// <summary>
        /// A sample regex pattern to use.
        /// </summary>
        private const string pattern = "http://topcoder.com/tco[0-9]+";

        /// <summary>
        /// A sample title string to use.
        /// </summary>
        private const string title = "Default title";

        /// <summary>
        /// A sample url to use.
        /// </summary>
        private const string url = "http://www.topcoder.com/";

        /// <summary>
        /// Tests constructor and property getter accuracy.
        /// </summary>
        [Test]
        public void TestCtorGetter()
        {
            RegexNodeMatcher matcher = new RegexNodeMatcher(title, url, pattern);

            Assert.AreEqual(pattern, matcher.NodePattern, "NodePattern is not initialized properly");
            Assert.AreEqual(title, matcher.Title, "The Title is not initialized properly");
            Assert.AreEqual(url, matcher.Url, "The Url is not initialized properly");
        }

        /// <summary>
        /// Tests constructor given null title.
        /// Expecting ArgumentNullException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestCtorNullTitle()
        {
            new RegexNodeMatcher(null, url, pattern);
        }

        /// <summary>
        /// Tests constructor given null url.
        /// Expecting ArgumentNullException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestCtorNullUrl()
        {
            new RegexNodeMatcher(title, null, pattern);
        }

        /// <summary>
        /// Tests constructor given null pattern.
        /// Expecting ArgumentNullException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestCtorNullPattern()
        {
            new RegexNodeMatcher(title, url, null);
        }

        /// <summary>
        /// Tests constructor given empty title.
        /// Expecting ArgumentException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestCtorEmptyTitle()
        {
            new RegexNodeMatcher(" ", url, pattern);
        }

        /// <summary>
        /// Tests constructor given empty url.
        /// Expecting ArgumentException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestCtorEmptyUrl()
        {
            new RegexNodeMatcher(title, " ", pattern);
        }

        /// <summary>
        /// Tests constructor given empty pattern.
        /// Expecting ArgumentException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestCtorEmptyPattern()
        {
            new RegexNodeMatcher(title, url, " ");
        }

        /// <summary>
        /// Tests constructor given invalid regex.
        /// Expecting ArgumentException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestCtorInvalidPattern()
        {
            new RegexNodeMatcher(title, url, ".+*");
        }

        /// <summary>
        /// Tests Matches accuracy for the sample pattern.
        /// </summary>
        [Test]
        public void TestMatches()
        {
            INodeMatcher matcher = new RegexNodeMatcher(title, url, pattern);

            Assert.IsTrue(matcher.Matches(new BreadCrumbNode(
                null, "http://topcoder.com/tco0")), "Digit should match");
            Assert.IsTrue(matcher.Matches(new BreadCrumbNode(
                null, "http://topcoder.com/tco12345")), "One or more digit is ok");
            Assert.IsFalse(matcher.Matches(new BreadCrumbNode(
                null, "http://topcoder.com/tcoxx")), "Character should not match");
            Assert.IsFalse(matcher.Matches(new BreadCrumbNode(
                null, "http://topcoder.com/tco")), "At least 1 digit is required");
            Assert.IsFalse(matcher.Matches(new BreadCrumbNode(
                null, "http://www.com/tco06")), "Different host");
        }

        /// <summary>
        /// Tests that input is considered a single line by the regex.
        /// </summary>
        [Test]
        public void TestMatchesMultiLine()
        {
            // matches only on beginning of the complete text
            INodeMatcher matcher = new RegexNodeMatcher(title, url, "^" + pattern);

            Assert.IsTrue(matcher.Matches(new BreadCrumbNode(
                null, "http://topcoder.com/tco06\r\nSecond line")), "String is treated as single line");
            Assert.IsFalse(matcher.Matches(new BreadCrumbNode(
                null, "First line\r\nhttp://topcoder.com/tco06")), "Cannot match in second line");
        }

        /// <summary>
        /// Tests Matches given null/uninitialized node (contains null Url).
        /// Expecting ArgumentNullException.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestMatchesNull()
        {
            new RegexNodeMatcher(title, url, pattern).Matches(
                new BreadCrumbNode());
        }

        /// <summary>
        /// Tests Equals given two RegexNodeMatcher with the same/different title, url, and pattern.
        /// </summary>
        [Test]
        public void TestEquals()
        {
            INodeMatcher m1 = new RegexNodeMatcher(title, url, pattern);
            INodeMatcher m2 = new RegexNodeMatcher(title, url, pattern);
            INodeMatcher m3 = new RegexNodeMatcher(title + " ", url, pattern);
            INodeMatcher m4 = new RegexNodeMatcher(title, "+" + url, pattern);
            INodeMatcher m5 = new RegexNodeMatcher(title, url, pattern + pattern);

            Assert.IsTrue(m1.Equals(m2), "Same object");
            Assert.IsFalse(m1.Equals(m3), "Different title");
            Assert.IsFalse(m1.Equals(m4), "Different url");
            Assert.IsFalse(m1.Equals(m5), "Different pattern");
        }

        /// <summary>
        /// Tests Equals return false given null object.
        /// </summary>
        [Test]
        public void TestEqualsNull()
        {
            Assert.IsFalse(new RegexNodeMatcher(title, url, pattern).Equals(null),
                "Null is different");
        }

        /// <summary>
        /// Tests Equals return false given different instance type.
        /// </summary>
        [Test]
        public void TestEqualsDifferentType()
        {
            // compare with MyNodeMatcher
            Assert.IsFalse(new RegexNodeMatcher(title, url, pattern).Equals(
                new MyNodeMatcher(title, url)),
                "Different instance type");
        }

        /// <summary>
        /// Tests GetHashCode that it returns the same code for the same object
        /// and different code for different object.
        /// </summary>
        [Test]
        public void TestGetHashCode()
        {
            INodeMatcher m1 = new RegexNodeMatcher(title, url, pattern);
            INodeMatcher m2 = new RegexNodeMatcher(title, url, pattern);
            INodeMatcher m3 = new RegexNodeMatcher(title + title, url, pattern);

            Assert.AreEqual(m1.GetHashCode(), m2.GetHashCode(), "Same object content");
            Assert.IsTrue(m1.GetHashCode() != m3.GetHashCode(), "Different object content");
        }

        /// <summary>
        /// Tests that the node matcher is serializable and deserializable.
        /// </summary>
        [Test]
        public void TestSerialization()
        {
            // serialize using BinaryFormatter to a MemoryStream
            INodeMatcher matcher = new RegexNodeMatcher(title, url, pattern);
            MemoryStream stream = new MemoryStream();
            BinaryFormatter formatter = new BinaryFormatter();
            formatter.Serialize(stream, matcher);

            // rewind the stream, and deserialize
            stream.Position = 0;
            object obj = formatter.Deserialize(stream);
            Assert.IsTrue(obj is RegexNodeMatcher, "RegexNodeMatcher is deserialized");

            // assert deserialized object
            RegexNodeMatcher mm = (RegexNodeMatcher) obj;
            Assert.IsFalse(matcher == mm, "Different instance");
            Assert.AreEqual(title, mm.Title, "The title is deserialized");
            Assert.AreEqual(url, mm.Url, "The url is deserialized");
            Assert.AreEqual(pattern, mm.NodePattern, "The node pattern is deserialized");

            // try matches using deserialized object
            Assert.IsTrue(mm.Matches(new BreadCrumbNode(
                null, "http://topcoder.com/tco06")), "Digit should match");
            Assert.IsFalse(mm.Matches(new BreadCrumbNode(
                null, "http://topcoder.com/tco")), "At least 1 digit is required");
        }
    }
}
